//  _2DCleanYUY2 Ver.0.10 : 2D Noise Cleaner for Avisynth
//  Copyright (C) 2002 kiraru2002
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//------------------------------------------------------------------------------------
//  This program is ported and modified to YUY2 color space by Kiraru2002(masani)
//  http://members.tripod.co.jp/kiraru2002/
//
// Original VirtualDub's plug-ins are "2D cleaner v0.6 by and Jim Casaburi"
// and "Optimized 2D cleaner for PIII ver 0.9 by Jaan Kalda".
// 
// You MUST NOT contact to Jim Casaburi,Jaan Kalda, Avery Lee and Donald A. Graft
// for any trouble which happened using this plug-in, and for any things about this plug-in.
//

/*
    2d cleaner filter for VirtualDub -- blends a pixel with pixels
	surrounding it as long as those pixels are simular to the source
	pixel

    Copyright (C) 2000 Jim Casaburi
		Based on code by Avery Lee
		Useful suggestions and much help from Donald A. Graft

	The author can be contacted at:
	Jim Casaburi
	casaburi@earthlink.net

	New versions of the source code and the compiled filter can be
	found at http://home.earthlink.net/~casaburi/download/
*/
//----------------------------------------------------------------------------------------------
//#define MAKEWRITABLE
#define AVISYNTH_2_5

#ifndef AVISYNTH_2_5
	#define	AVISYNTH_H	"avisynth.h"
	#define	AVISYNTHPLUGININIT	AvisynthPluginInit
	typedef int			__INT64;
#else
	#define AVISYNTH_H	"avisynth2.h"
	#define	AVISYNTHPLUGININIT	AvisynthPluginInit2
	typedef __int64		__INT64;
#endif

#ifdef _MSC_VER
	#include <crtdbg.h>
	#if _MSC_VER == 1100	//MSC5.0 ?
		#define noreturn
	#endif
#else
	#define _ASSERTE(x) assert(x)
	#include <assert.h>
#endif


#include <stdio.h>
#include <windows.h>
#include <math.h>
#include <malloc.h>
#include AVISYNTH_H

extern "C" int __stdcall ppixels(unsigned __int64		Thresh,
								 unsigned long			*srcp,
								 unsigned long			*dstp,
								 unsigned long			*aligned_YUVsrcp,
								 unsigned long			*aligned_YUVdstp,
								 int					fDebugNoise,
								 int					radiusX,
								 int					radiusY,
								 int					multfact,
								 int					pitchm,
								 int					width,
								 int					hight,
								 int					src_modulo,
								 int					dst_modulo
						 );


extern "C"	unsigned __int64 divtable[128] ={0}, n_by_2[128] = {0}; //7*18

#ifdef _DEBUG
extern "C"	unsigned __int64 debugblock[720 * 480] = {0};
#endif

class TwoDClean : public GenericVideoFilter
{
	PVideoFrame src,dst;
	int		thresholdY, radiusX, radiusY;
	int		thresholdU,thresholdV;
	int		fDebugNoise,fInterlace;
	int		pos_Y;
	int		multfact,pitch,pitchm;
	unsigned __int64 Thresh;
	int		width,height;
	unsigned long *srcp;
	unsigned long  *dstp;
	int		src_modulo, dst_modulo;
	
	VideoInfo	tmp_vi;
	unsigned long *YUVsrcp,*YUVdstp,*aligned_YUVsrcp,*aligned_YUVdstp;

public:
	TwoDClean::TwoDClean(PClip _child, int _mode, int _thresholdY,
						int _radiusX, int _radiusY,
						int _debugnoise, int _thresholdU, int _thresholdV,
						IScriptEnvironment* env) :
					GenericVideoFilter(_child),
					fInterlace(_mode),thresholdY(_thresholdY), radiusX(_radiusX), radiusY(_radiusY),
					fDebugNoise(_debugnoise),thresholdU(_thresholdU),thresholdV(_thresholdV)
	{
		TwoDClean_Init(env);

		if (!vi.IsYUY2())
			env->ThrowError("_2DCleanYUY2: requires YUY2 input.");

		if (radiusX < 0 || radiusX > 3)
			env->ThrowError("_2DCleanYUY2: parameter error.(radiusX range is 0 to 3)");

		if (radiusY < 0 || radiusY > 9)
			env->ThrowError("_2DCleanYUY2: parameter error.(radiusY range is 0 to 9)");

		if (fDebugNoise < 0 || fDebugNoise > 4)
			env->ThrowError("_2DCleanYUY2: parameter error.(dmode range is 0 to 4)");
	}

	TwoDClean::~TwoDClean()
	{
		if(YUVsrcp)		{ free(YUVsrcp); YUVsrcp = NULL; }
	}

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);

private:
	void TwoDClean_Init(IScriptEnvironment* env);

};


PVideoFrame __stdcall TwoDClean::GetFrame(int frame, IScriptEnvironment* env)
{
	int rc;

#ifdef MAKEWRITABLE
	src = child->GetFrame(frame, env);
	env->MakeWritable(&src);
	pitch = pitchm = src->GetPitch();
	width = src->GetRowSize()>>1;
	height = src->GetHeight();
	src_modulo = pitch - src->GetRowSize();
	srcp = (unsigned long *)src->GetReadPtr();
	if (fInterlace)		pitchm <<= 1;

	dstp = (unsigned long *)src->GetWritePtr();
#else
	src = child->GetFrame(frame, env);
	pitch = pitchm = src->GetPitch();
	width = src->GetRowSize()>>1;
	height = src->GetHeight();
	src_modulo = pitch - src->GetRowSize();
	srcp = (unsigned long *)src->GetReadPtr();
	if (fInterlace)		pitchm <<= 1;

	dst = env->NewVideoFrame(vi, 16);
	dstp = (unsigned long *)dst->GetWritePtr();
	dst_modulo = dst->GetPitch() - dst->GetRowSize();
#endif

	rc = ppixels(	Thresh,
					srcp,
					dstp,
					aligned_YUVsrcp,
					aligned_YUVdstp,
					fDebugNoise,
					radiusX,
					radiusY,
					multfact,
					pitchm,
					width,
					height-1,
					src_modulo,
#ifdef MAKEWRITABLE
					src_modulo
#else
					dst_modulo
#endif
					);
	
#ifdef MAKEWRITABLE
	return	src;
#else
	return	dst;
#endif
}

void TwoDClean::TwoDClean_Init(IScriptEnvironment* env)
{
	int					i;
	unsigned __int64	u;

	//here we initialize the table, which allows us to substitute divisions with multiplication
	for (i=2; i<128; ++i){
		u = (unsigned __int64)((double)(0x10000) / i + 0.999);
		divtable[i] = u | (u<<16) | (u<<32) | (u<<48);
		n_by_2[i] = (i/2) | (i/2)<<16 | (i/2)<<32 | (i/2)<<48;
	}
	divtable[1] = 0xffffffffffffffffi64;
	n_by_2[1]   = 0x0001000100010001i64;

	src = child->GetFrame(0, env);
	width = src->GetRowSize()>>1;
	height = src->GetHeight();
	if (fInterlace) {
		multfact = 2;
	} else {
		multfact = 1;
	}

	if (thresholdU==-1) thresholdU=thresholdY;
	if (thresholdV==-1) thresholdV=thresholdY;
	Thresh = (unsigned __int64)(thresholdY)
			| ((unsigned __int64)thresholdU << 16)
			| ((unsigned __int64)(thresholdY)<< 32)
			| ((unsigned __int64)thresholdV << 48);

	if((YUVsrcp = (unsigned long *)malloc(width * 4 * height + 16))==NULL) 
	{
		env->ThrowError("_2DCleanYUY2: not enough memory.");
	}

	aligned_YUVsrcp = (unsigned long *)(((unsigned long)YUVsrcp + 15) & -16);
	return;
}



AVSValue __cdecl Create_2DCleanYUY2Opt(AVSValue args, void* user_data, IScriptEnvironment* env) {
    return new TwoDClean(args[0].AsClip(),
						args[1].AsInt(0),		// Mode bit0:interlace bit1:edge
						args[2].AsInt(9),		// ThresholdY
						args[3].AsInt(2),		// RadiusX
						args[4].AsInt(2),		// RadiusY
						args[5].AsInt(0),		// DebugNoise
						args[6].AsInt(-1),		// ThresholdU
						args[7].AsInt(-1),		// ThresholdV
						env);
}

extern "C" __declspec(dllexport) const char* __stdcall AVISYNTHPLUGININIT(IScriptEnvironment* env)
{
    env->AddFunction("_2DCleanYUY2",
					 "c[interlaced]i[thresholdY]i[radiusX]i[radiusY]i" \
					 "[dmode]i[thresholdU]i[thresholdV]i",
		             Create_2DCleanYUY2Opt, 0);
	return 0;
}
